#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from kiwi3d.halfedges import halfedges
from kiwi3d.array_correspondence import array_correspondence

def triangle_triangle_adjacency(F):
    """Given a manifold triangle mesh with face indices F, computes adjacency
    info between triangles using PyTorch
    
    The ordering convention for halfedges is the following:
    [halfedge opposite vertex 0,
     halfedge opposite vertex 1,
     halfedge opposite vertex 2]

    Parameters
    ----------
    F : (m,3) torch int tensor
        face index list of a triangle mesh

    Returns
    -------
    TT : (m,3) torch int tensor
        Index list specifying which face j is adjacent to which face i across
        the respective halfedge in position (i,j).
        If there is no adjacent face (boundary halfedge), the entry is -1.
    TTi : (m,3) torch int tensor
        Index list specifying which halfedge of face j (0,1,2) is adjacent to i
        in position (i,j).

    Examples
    --------
    ```python
    from gpytoolbox import regular_square_mesh, triangle_triangle_adjacency
    v, f = regular_square_mesh(10)
    TT, TTi = triangle_triangle_adjacency(f)
    ```
    """
    
    m = F.shape[0]
    assert m > 0
    assert F.shape[1] == 3

    he = halfedges(F)
    he_flat = torch.cat((he[:,0,:], he[:,1,:], he[:,2,:]), dim=0)
    he_flip_flat = torch.flip(he_flat, dims=[1])

    map_to_flip = array_correspondence(he_flat, he_flip_flat, axis=0)
    TT = torch.where(map_to_flip<0, -1, map_to_flip % m).reshape(F.T.shape).T.reshape(F.shape)
    TTi = torch.where(map_to_flip<0, -1, map_to_flip // m).reshape(F.T.shape).T.reshape(F.shape)

    return TT, TTi

if __name__ == '__main__':
    from gpytoolbox import regular_square_mesh
    v, f = regular_square_mesh(10) 
    TT_pt, TTi_pt = triangle_triangle_adjacency(torch.from_numpy(f))




import torch

def halfedges(F):
    """Given a triangle mesh with face indices F, returns all oriented halfedges
    as indices into the vertex array using PyTorch.

    The ordering convention for halfedges is the following:
    [halfedge opposite vertex 0,
     halfedge opposite vertex 1,
     halfedge opposite vertex 2]

    Parameters
    ----------
    F : (m,3) torch int tensor
        face index list of a triangle mesh

    Returns
    -------
    he : (m,3,2) torch int tensor
        halfedge list as per above conventions

    Examples
    --------
    ```python
    # Sample mesh
    v = torch.tensor([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
    f = torch.tensor([[0, 1, 2]], dtype=torch.int)
    # Call to halfedges
    he = halfedges(f)
    ```
    """
    
    assert F.shape[0] > 0
    assert F.shape[1] == 3

    he = torch.cat([F[:, [1, 2]].unsqueeze(1),
                    F[:, [2, 0]].unsqueeze(1),
                    F[:, [0, 1]].unsqueeze(1)], dim=1)
    
    return he

if __name__ == '__main__':
    # Example usage
    v = torch.tensor([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
    f = torch.tensor([[0, 1, 2]], dtype=torch.int)
    he = halfedges(f)
    print(he)  # Should match the expected halfedge list





import torch

def unique(x, return_inverse=False, dim=-1):
    unique, inverse = torch.unique(x, return_inverse=True, dim=dim)
    perm = torch.arange(inverse.size(dim), dtype=inverse.dtype, device=inverse.device)
    inverse_, perm = inverse.flip([dim]), perm.flip([dim])
    if return_inverse:
        return unique, inverse_.new_empty(unique.size(dim)).scatter_(dim, inverse_, perm), inverse

    return unique, inverse_.new_empty(unique.size(dim)).scatter_(dim, inverse_, perm)

def array_correspondence(A, B, axis=None):
    """Computes a map from A to the equal elements in B

    Parameters
    ----------
    A : (a,) or (a,k) torch tensor (must be 1-dim or 2-dim)
    B : (b,) or (b,k) torch tensor (must be 1-dim or 2-dim)
    axis : int or None, optional (default None)
        If None, will treat A and B as flat arrays.
        If a number, will check for equality of the entire axis, in which case
        the dimension of A and B across that axis must be equal.

    Returns
    -------
    f : (a,) torch long tensor 
        index list mapping from A to B, with -1 if there is no
        matching entry.
        If b contains multiple eligible entries, return an arbitrary one.
        If there are no -1s, `b[f] == a`

    Examples
    --------
    Example with simple array correspondence:
    ```python
    >>> A = np.array([1,7,2,7,9,3,7,0])
    >>> B = np.array([7,7,3,2,7,8])
    >>> f = gpy.array_correspondence(A, B)
    >>> f
    array([-1,  0,  3,  0, -1,  2,  0, -1])
    ```
    Example with row correspondence:
    ```python
    >>> A = np.array([[1,3],[5,2],[1,2],[5,2]])
    >>> B = np.array([[1,2],[6,9],[5,2]])
    >>> f = gpy.array_correspondence(A, B, axis=0)
    >>> f
    array([-1,  2,  0,  2])
    ```
    
    """
    if axis not in (None, 0, 1, -1, -2):
        raise Exception("Axis can only be None, 0, 1, -1, -2")
    if len(A.shape) > 2 or len(B.shape) > 2:
        raise Exception("Inputs A, B can only be up to 2 dimensional")

    if axis == 1 or axis == -1:
        A = A.transpose(0, 1)
        B = B.transpose(0, 1)
        axis = 0

    if axis is None:
        A = A.flatten()
        B = B.flatten()
        axis = 0

    # Convert tensors to contiguous memory
    A = A.contiguous()
    B = B.contiguous()

    # Get unique elements and their first occurrence indices
    uB, mapB = unique(B, dim=axis)

    # Concatenate unique B with A
    concatenated = torch.cat((uB, A), dim=axis)

    # Get unique elements of concatenated array and their inverse indices
    _, idx, inv = unique(concatenated, return_inverse=True, dim=axis)

    # Calculate the mapping
    imap = idx[inv[uB.shape[0]:]]
    imap[imap>=uB.shape[0]] = -1
    f = torch.where(imap < 0, torch.tensor(-1, dtype=torch.long), mapB[imap])

    return f

if __name__ == '__main__':
    A = torch.Tensor([1, 7, 2, 7, 9, 3, 7, 0])
    B = torch.Tensor([7, 7, 3, 2, 7, 8])
    f = array_correspondence(A, B)
    print(f)




import torch

def per_face_normals(V, F, unit_norm=True):
    """Vector perpedicular to all faces on a mesh using PyTorch
    
    Computes per face (optionally unit) normal vectors for a triangle mesh.

    Parameters
    ----------
    V : (n,d) torch tensor
        vertex list of a triangle mesh
    F : (m,d) torch int tensor
        face index list of a triangle mesh
    unit_norm : bool, optional (default True)
        Whether to normalize each face's normal before outputting

    Returns
    -------
    N : (n,d) torch double tensor
        Matrix of per-face normals

    See Also
    --------
    per_vertex_normals.
    """     

    dim = V.shape[1]

    if dim == 2:
        # Edge vectors
        v0 = V[F[:, 0], :]
        v1 = V[F[:, 1], :]
        # Difference between edge vectors
        e = v1 - v0
        # Rotate by 90 degrees
        N = torch.cat((e[:, 1][:, None], -e[:, 0][:, None]), dim=1)
    elif dim == 3:     
        v0 = V[F[:, 0], :]
        v1 = V[F[:, 1], :]
        v2 = V[F[:, 2], :]

        # It's basically just a cross product
        N = torch.cross(v1 - v0, v2 - v0, dim=1)

    if unit_norm:
        N = N / N.norm(dim=1, keepdim=True)

    return N

if __name__ == '__main__':
    from gpytoolbox import read_mesh
    v, f = read_mesh('./spot_triangulated.obj')
    n_pt = per_face_normals_pt(torch.from_numpy(v), torch.from_numpy(f))

